/* $Id$

Copyright (C) 2010  The PARI group.

This file is part of the PARI/GP package.

PARI/GP is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation. It is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY WHATSOEVER.

Check the License for details. You should have received a copy of it, along
with the package; see the file 'COPYING'. If not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
#include "pari.h"
#include "paripriv.h"
#include <pthread.h>

struct mt_queue
{
  long no;
  GEN input, output;
  GEN worker;
  pthread_cond_t cond, cond1, cond2;
  pthread_mutex_t mut, mut1, mut2;
  pthread_cond_t *pcond;
  pthread_mutex_t *pmut;
};

struct mt_state
{
  pthread_t *th;
  struct pari_thread *pth;
  struct mt_queue *mq;
  long n, nbint, last;
  long pending;
  pthread_cond_t pcond;
  pthread_mutex_t pmut;
};

static THREAD long mt_thread_no = -1;
static struct mt_state *pari_mt;
void mt_init_stack(size_t s) { }

#define LOCK(x) pthread_mutex_lock(x); do
#define UNLOCK(x) while(0); pthread_mutex_unlock(x)

void pari_mt_init(void)
{
  pari_mt = NULL;
#ifdef _SC_NPROCESSORS_CONF
  pari_mt_nbthreads = sysconf(_SC_NPROCESSORS_CONF);
#else
  pari_mt_nbthreads = 1;
#endif
}

static void*
mt_queue_run(void *arg)
{
  GEN args = pari_thread_start((struct pari_thread*) arg);
  struct mt_queue *mq = (struct mt_queue *) args;
  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS,NULL);
  mt_thread_no = mq->no;
  for(;;)
  {
    GEN work, done;
    LOCK(&mq->mut)
    {
      while(!mq->input)
        pthread_cond_wait(&mq->cond, &mq->mut);
    } UNLOCK(&mq->mut);
    work = gcopy(mq->input);
    LOCK(&mq->mut1)
    {
      mq->input = NULL;
      pthread_cond_signal(&mq->cond1);
    } UNLOCK(&mq->mut1);
    done = closure_callgenvec(mq->worker,work);
    LOCK(mq->pmut)
    {
      mq->output = done;
      pthread_cond_signal(mq->pcond);
    } UNLOCK(mq->pmut);
    LOCK(&mq->mut2)
    {
      while(mq->output)
        pthread_cond_wait(&mq->cond2, &mq->mut2);
    } UNLOCK(&mq->mut2);
  }
  return NULL;
}

void *
mt_queue_start(GEN worker)
{
  long NBT = pari_mt_nbthreads;
  struct mt_state *mt = pari_malloc(sizeof(struct mt_state));
  long mtparisize = (long) (top-bot);
  long i;
  if (pari_mt) pari_err(e_MISC,"pending thread still exist");
  mt->mq     = pari_malloc(sizeof(*mt->mq) *NBT);
  mt->th     = pari_malloc(sizeof(*mt->th) *NBT);
  mt->pth    = pari_malloc(sizeof(*mt->pth)*NBT);
  mt->pending = 0;
  mt->n = NBT;
  mt->nbint = 0;
  mt->last = 0;
  pthread_cond_init(&mt->pcond,NULL);
  pthread_mutex_init(&mt->pmut,NULL);

  for (i=0;i<NBT;i++)
  {
    struct mt_queue *mq = mt->mq+i;
    mq->no     = i;
    mq->worker = worker;
    mq->input  = NULL;
    mq->output = NULL;
    mq->pcond  = &mt->pcond;
    mq->pmut   = &mt->pmut;
    pthread_cond_init(&mq->cond,NULL);
    pthread_cond_init(&mq->cond1,NULL);
    pthread_cond_init(&mq->cond2,NULL);
    pthread_mutex_init(&mq->mut,NULL);
    pthread_mutex_init(&mq->mut1,NULL);
    pthread_mutex_init(&mq->mut2,NULL);
    pari_thread_alloc(&mt->pth[i],mtparisize,(void*)mq);
  }
  if (DEBUGLEVEL) pari_warn(warner,"start threads");
  for (i=0;i<NBT;i++)
    pthread_create(&mt->th[i],NULL, &mt_queue_run, (void*)&mt->pth[i]);
  pari_mt = mt;
  return mt;
}


static long
mt_queue_check(struct mt_state *mt)
{
  long i;
  for(i=0; i<mt->n; i++)
  {
    struct mt_queue *mq = mt->mq+i;
    if (mq->output) return i;
  }
  return -1;
}

GEN
mt_queue_get(void *p, long *pending)
{
  GEN done = NULL;
  struct mt_state *mt = (struct mt_state *) p;
  struct mt_queue *mq;
  long last;
  if (mt->nbint<mt->n) { mt->last = mt->nbint; *pending = mt->pending; return NULL; }
  LOCK(&mt->pmut)
  {
    while ((last = mt_queue_check(mt)) < 0)
      pthread_cond_wait(&mt->pcond, &mt->pmut);
  } UNLOCK(&mt->pmut);
  mq = mt->mq+last;
  if (mq->output==gnil)   pari_err(e_STACK);
  done = gcopy(mq->output);
  if (typ(done)==t_ERROR) pari_err(0,done);
  LOCK(&mq->mut2)
  {
    mq->output = NULL;
    pthread_cond_signal(&mq->cond2);
  } UNLOCK(&mq->mut2);
  mt->last = last;
  mt->pending--;
  *pending = mt->pending;
  return done;
}

void
mt_queue_submit(void *p, GEN work)
{
  struct mt_state *mt = (struct mt_state *) p;
  struct mt_queue *mq = mt->mq+mt->last;
  if (!work) { mt->nbint=mt->n; return; }
  if (mt->nbint<mt->n) mt->nbint++;
  LOCK(&mq->mut)
  {
    mq->input = work;
    pthread_cond_signal(&mq->cond);
  } UNLOCK(&mq->mut);
  mt->pending++;
  LOCK(&mq->mut1)
  {
    while (mq->input)
      pthread_cond_wait(&mq->cond1, &mq->mut1);
  } UNLOCK(&mq->mut1);
}

void
mt_queue_end(void *p)
{
  struct mt_state *mt = (struct mt_state *) p;
  long i;
  void *res;
  for (i=0; i<mt->n; i++)
    pthread_cancel(mt->th[i]);
  for (i=0; i<mt->n; i++)
    pthread_join(mt->th[i],(void*)&res);
  if (DEBUGLEVEL) pari_warn(warner,"stop threads");
  pari_mt = NULL;
  for (i=0;i<mt->n;i++)
  {
    struct mt_queue *mq = mt->mq+i;
    pthread_cond_destroy(&mq->cond);
    pthread_cond_destroy(&mq->cond1);
    pthread_cond_destroy(&mq->cond2);
    pthread_mutex_destroy(&mq->mut);
    pthread_mutex_destroy(&mq->mut1);
    pthread_mutex_destroy(&mq->mut2);
    pari_thread_free(&mt->pth[i]);
  }
  pari_free(mt->mq);
  pari_free(mt->pth);
  pari_free(mt->th);
  pari_free(p);
}

void
mt_err_recover(long er)
{
  if (mt_thread_no>=0)
  {
    struct mt_state *mt = pari_mt;
    struct mt_queue *mq = mt->mq+mt_thread_no;
    LOCK(mq->pmut)
    {
      mq->output = global_err_data ? global_err_data : gnil;
      pthread_cond_signal(mq->pcond);
    } UNLOCK(mq->pmut);
    pthread_exit((void*)1);
  }
  else if (pari_mt) mt_queue_end(pari_mt);
}

void
pari_mt_close(void) { }
